from logging import Logger
import json
import os
from typing import Any, Callable, Dict, List, Optional

import pandas as pd
import torch

from mu_map.dataset.default import MuMapDataset
from mu_map.dataset.normalization import (
    GaussianNormTransform,
    MaxNormTransform,
    MeanNormTransform,
    norm_by_str,
)
from mu_map.dataset.patches import MuMapPatchDataset
from mu_map.dataset.transform import PadCropTranform, SequenceTransform
from mu_map.eval.measures import compute_measures
from mu_map.models.discriminator import Discriminator, PatchDiscriminator
from mu_map.models.unet import UNet
from mu_map.training.cgan import cGANTraining, DiscriminatorParams, GeneratorParams
from mu_map.training.loss import WeightedLoss
from mu_map.random_search.lib import RandomSearch
from mu_map.random_search.sampling import (
    ChoiceSampler,
    DependentChoiceSampler,
    IntIntervalSampler,
    FloatIntervalSampler,
)


class cGANRandomSearch(RandomSearch):
    """
    Implementation of a random search for cGAN training.
    """

    def __init__(
        self,
        iterations: int,
        device: torch.device,
        out_dir: str,
        train_dir: str = "train_data",
        dataset_dir: str = "data/second",
        logger: Optional[Logger] = None,
    ):
        super().__init__(
            param_sampler={},
            iterations=iterations,
            device=device,
            out_dir=out_dir,
            train_dir=train_dir,
            logger=logger,
        )

        self.n_slices = 32

        def patch_number(patch_size: int, **kwargs):
            if patch_size == 32:
                return list(range(50, 101))
            else:
                return list(range(25, 51))

        # dataset parameters
        self.param_sampler["patch_offset"] = ChoiceSampler([0])
        self.param_sampler["patch_size"] = ChoiceSampler([32, 64])
        self.param_sampler["patch_number"] = DependentChoiceSampler(patch_number)
        self.param_sampler["scatter_correction"] = ChoiceSampler([False])
        self.param_sampler["shuffle"] = ChoiceSampler([False, True])
        self.param_sampler["normalization"] = ChoiceSampler(
            [MeanNormTransform(), MaxNormTransform(), GaussianNormTransform()]
        )
        self.param_sampler["pad_crop"] = ChoiceSampler(
            [None, PadCropTranform(dim=3, size=self.n_slices)]
        )

        # model parameters
        self.param_sampler["discriminator_type"] = ChoiceSampler(["class", "patch"])

        def discriminator_conv_features(discriminator_type: str, **kwargs):
            if discriminator_type == "class":
                return [[32, 64, 128], [64, 128, 256], [32, 64, 128, 256]]
            else:
                return [[32, 64, 128, 256], [64, 128, 256, 512]]

        self.param_sampler["discriminator_conv_features"] = DependentChoiceSampler(
            discriminator_conv_features
        )
        self.param_sampler["generator_features"] = ChoiceSampler(
            [
                [64, 128, 256, 512],
                [32, 64, 128, 256, 512],
            ]
        )

        # training parameters
        self.param_sampler["epochs"] = ChoiceSampler([100])

        def batch_size(patch_size: int, **kwargs):
            return [32] if patch_size == 64 else [64]

        # self.param_sampler["batch_size"] = ChoiceSampler([32, 64])
        self.param_sampler["batch_size"] = DependentChoiceSampler(batch_size)
        # self.param_sampler["lr"] = FloatIntervalSampler(0.1, 0.0001)
        self.param_sampler["lr"] = ChoiceSampler([0.001])
        self.param_sampler["lr_decay"] = ChoiceSampler([False, True])
        self.param_sampler["lr_decay_epoch"] = ChoiceSampler([1])
        self.param_sampler["lr_decay_factor"] = ChoiceSampler([0.99])
        self.param_sampler["criterion_dist"] = ChoiceSampler(
            [WeightedLoss.from_str("L1"), WeightedLoss.from_str("L2+GDL")]
        )
        self.param_sampler["weight_criterions"] = ChoiceSampler(
            [
                (100.0, 1.0),
                (20.0, 1.0),
                (5.0, 1.0),
                (1.0, 1.0),
                (1.0, 5.0),
                (1.0, 20.0),
                (1.0, 100.0),
            ]
        )

    def setup_training(
        self, params: Dict[str, Any], train_dir: str, snapshot_dir: str, logger: Logger
    ) -> cGANTraining:
        transforms = [params["normalization"], params["pad_crop"]]
        transforms = list(filter(lambda transform: transform is not None, transforms))
        transform_normalization = SequenceTransform(transforms)

        dataset = MuMapPatchDataset(
            self.dataset_dir,
            patches_per_image=params["patch_number"],
            patch_size=params["patch_size"],
            patch_size_z=self.n_slices,
            patch_offset=params["patch_offset"],
            shuffle=params["shuffle"],
            transform_normalization=transform_normalization,
            scatter_correction=params["scatter_correction"],
            logger=logger,
        )

        input_size = (
            self.n_slices,
            params["patch_size"],
            params["patch_size"],
        )
        if params["discriminator_type"] == "class":
            discriminator = Discriminator(
                in_channels=2,
                input_size=input_size,
                conv_features=params["discriminator_conv_features"],
            )
        else:
            discriminator = PatchDiscriminator(
                in_channels=2, features=params["discriminator_conv_features"]
            )
        discriminator = discriminator.to(self.device)

        optimizer = torch.optim.Adam(
            discriminator.parameters(), lr=params["lr"], betas=(0.5, 0.999)
        )
        lr_scheduler = (
            torch.optim.lr_scheduler.StepLR(
                optimizer,
                step_size=params["lr_decay_epoch"],
                gamma=params["lr_decay_factor"],
            )
            if params["lr_decay"]
            else None
        )
        params_d = DiscriminatorParams(
            model=discriminator, optimizer=optimizer, lr_scheduler=lr_scheduler
        )

        generator = UNet(in_channels=1, features=params["generator_features"])
        generator = generator.to(self.device)
        optimizer = torch.optim.Adam(
            generator.parameters(), lr=params["lr"], betas=(0.5, 0.999)
        )
        lr_scheduler = (
            torch.optim.lr_scheduler.StepLR(
                optimizer,
                step_size=params["lr_decay_epoch"],
                gamma=params["lr_decay_factor"],
            )
            if params["lr_decay"]
            else None
        )
        params_g = GeneratorParams(
            model=generator, optimizer=optimizer, lr_scheduler=lr_scheduler
        )

        return cGANTraining(
            epochs=params["epochs"],
            dataset=dataset,
            batch_size=params["batch_size"],
            device=self.device,
            snapshot_dir=snapshot_dir,
            snapshot_epoch=params["epochs"],
            logger=logger,
            params_generator=params_g,
            params_discriminator=params_d,
            loss_func_dist=params["criterion_dist"],
            weight_criterion_dist=params["weight_crit_dist"],
            weight_criterion_adv=params["weight_crit_adv"],
            # early_stopping=10,
        )

    def eval_run(self) -> float:
        torch.set_grad_enabled(False)

        weights_file = os.path.join(self.training.snapshot_dir, "val_min_generator.pth")
        self.logger.debug(f"Load weights from {weights_file}")
        model = self.training.generator.eval()
        model.load_state_dict(torch.load(weights_file, map_location=self.device))

        transform_normalization = SequenceTransform(
            [self.params["normalization"], PadCropTranform(dim=3, size=self.n_slices)]
        )
        dataset = MuMapDataset(
            self.dataset_dir,
            split_name="validation",
            transform_normalization=transform_normalization,
            scatter_correction=self.params["scatter_correction"],
        )

        values = compute_measures(dataset, model)
        values.to_csv(os.path.join(self.train_dir, "measures.csv"), index=False)
        torch.set_grad_enabled(True)
        return values["NMAE"].mean()

    def get_score(self, train_dir: str) -> float:
        file_measures = os.path.join(train_dir, "measures.csv")
        measures = pd.read_csv(file_measures)
        return measures["NMAE"].mean()

    def cleanup_params(self, params: Dict[str, Any]) -> Dict[str, Any]:
        weight_crit_dist, weight_crit_adv = params["weight_criterions"]
        params["weight_crit_dist"] = weight_crit_dist
        params["weight_crit_adv"] = weight_crit_adv
        del params["weight_criterions"]
        return params


class ParamJSONDecoder(json.JSONDecoder):
    """
    A custom JSON decoder to the parameters of a cGAN random search run.
    """

    def __init__(self):
        super().__init__()

        self.int_fields = [
            "patch_size",
            "patch_offset",
            "patch_number",
            "epochs",
            "batch_size",
            "lr_decay_epoch",
        ]
        self.float_fields = [
            "lr",
            "lr_decay_factor",
            "weight_crit_dist",
            "weight_crit_adv",
        ]
        self.bool_fiels = ["scatter_correction", "shuffle", "lr_decay"]
        self.int_list_fields = ["discriminator_conv_features", "generator_features"]

    def decode(self, s: str) -> Dict[str, Any]:
        """
        Decode a JSON string into a dict of parameters.
        """
        params = super().decode(s)

        self.parse_fields(params, int, *self.int_fields)
        self.parse_fields(params, float, *self.float_fields)
        self.parse_fields(params, lambda v: v == "True", *self.bool_fiels)
        self.parse_fields(
            params, lambda v: self.parse_as_list(v, int), *self.int_list_fields
        )
        self.parse_fields(params, WeightedLoss.from_str, "criterion_dist")
        self.parse_fields(params, norm_by_str, "normalization")
        params["pad_crop"] = (
            None if params["pad_crop"] == "None" else PadCropTranform(dim=3, size=32)
        )

        return params

    def parse_fields(
        self, params: Dict[str, Any], func: Callable[str, Any], *fields: str
    ):
        """
        Parse fields in a dict with a specified function.

        This function makes sure that the fields exist and are currently string.

        Parameters
        ----------
        params: Dict[str, Any]
            the dict whose values are parsed
        func: Callable[str, Any]
            the function used for parsing
        *fields: str
            the fields to parse
        """
        fields = filter(lambda field: field in params.keys(), fields)
        fields = filter(lambda field: type(params[field]) == str, fields)
        for field in fields:
            params[field] = func(params[field])

    def parse_as_list(self, s: str, func: Callable[str, Any]) -> List[Any]:
        """
        Parse a field as a list.

        Parameters
        ----------
        s: str
            the string to be parsed as a list
        func:
            the parsing function for list elements

        Returns
        -------
        List[Any]
        """
        s = s[1:-1]  # remove brackets
        values = s.split(",")
        values = map(lambda x: x.strip(), values)
        values = map(func, values)
        return list(values)


def load_params(filename: str) -> Dict[str, Any]:
    """
    Load parameters of a cGAN random search from a file.

    Parameters
    ----------
    filename: str
        the file to be read

    Returns
    -------
    Dict[str, Any]
    """
    with open(filename, mode="r") as f:
        return json.load(f, cls=ParamJSONDecoder)


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(
        description="Perform random search hyperparameter optimization for cGAN training",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "--out_dir",
        type=str,
        default="cgan_random_search",
        help="the directory where results are stored",
    )
    parser.add_argument(
        "--iterations",
        type=int,
        default=10,
        help="number of random search iterations to perform",
    )
    parser.add_argument(
        "--device",
        type=str,
        default="cuda:0" if torch.cuda.is_available() else "cpu",
        help="the device (cpu or gpu) with which the training is performed",
    )

    args = parser.parse_args()
    device = torch.device(args.device)

    random_search = cGANRandomSearch(args.iterations, device, args.out_dir)
    random_search.run()
