"""
Implementation of random search for hyper parameter optimization.
"""
import json
from logging import Logger
import os
import shutil
import sys
from typing import Any, Dict, Optional, Tuple

import torch

from mu_map.logging import get_logger
from mu_map.random_search.sampling import ParamSampler, DependentChoiceSampler
from mu_map.training.lib import AbstractTraining, init_random_seed


class RandomSearch:
    """
    Abstract implementation of a random search procedure.

    Parameters
    ----------
    sampler: Dict[str, ParamSampler]
        a dict of name and parameter sampler pairs all of them are
        sampled for a single random search run
    iterations: int
        the number of random search iterations to perform
    device: torch.device
        the device on which training should be performed
    out_dir: str
        direcotry where results are stored
    train_dir: str
        dataset under `out_dir` where the results of the current training are stored
    dataset_dir: str
        directory of the dataset used for training
    logger: Logger
    """

    def __init__(
        self,
        param_sampler: Dict[str, ParamSampler],
        iterations: int,
        device: torch.device,
        out_dir: str,
        train_dir: str = "train_data",
        dataset_dir: str = "data/second",
        logger: Optional[Logger] = None,
    ):
        self.param_sampler = param_sampler
        self.iterations = iterations
        self.device = device
        self.out_dir = out_dir
        self.train_dir = os.path.join(self.out_dir, train_dir)
        self.dataset_dir = dataset_dir

        self.logger = (
            logger
            if logger is not None
            else get_logger(
                logfile=os.path.join(self.out_dir, "random_search.log"),
                loglevel="INFO",
                name=self.__class__.__name__,
            )
        )

        if not os.path.exists(self.out_dir):
            os.mkdir(self.out_dir)
        self.current_run, self.current_score = self.retrieve_state()

    def retrieve_state(self) -> Tuple[int, float]:
        """
        Retrieve the current state of for continuation of a random search procedure.

        Returns
        -------
        int
            the current run with which the random search should continue
        float
            the currently best score
        """
        runs = os.listdir(self.out_dir)
        runs = filter(lambda f: os.path.isdir(os.path.join(self.out_dir, f)), runs)
        runs = filter(lambda f: not os.path.islink(os.path.join(self.out_dir, f)), runs)
        runs = list(map(int, runs))

        if len(runs) == 0:
            return 1, sys.maxsize

        current_run = max(runs)
        current_score = self.get_score(os.path.join(self.out_dir, "best"))
        self.logger.info(
            f"Continue random search from run {current_run} and minimal score {current_score}"
        )
        return current_run + 1, current_score

    def sample(self) -> Dict[str, Any]:
        """
        Sample all parameters.

        This method makes sure that all dependent choice samplers get their
        required dependencies (need to be registered in order).

        Returns
        -------
        Dict[str, Any]
            a dictionary of name and drawn parameter pairs.
        """
        _params = {}
        for key, sampler in self.param_sampler.items():
            if isinstance(sampler, DependentChoiceSampler):
                param = sampler.sample(_params)
            else:
                param = sampler.sample()
            _params[key] = param
        return _params

    def serialize_params(
        self, params: Dict[str, Any], filename: Optional[str] = None
    ) -> str:
        """
        Serialize a set of parameters to json and dump them into a file.

        Parameters
        ----------
        params: Dict[str, Any]
            dict of params to be serialized
        filename: str, optional
            filename where the json is dumped

        Returns
        -------
        str
            the params as a json representation
        """
        _params = {}
        for key, value in params.items():
            _params[key] = str(value).replace("\n", "").replace("    ", ", ")

        _str = json.dumps(_params, indent=2)
        if filename is not None:
            with open(filename, mode="w") as f:
                f.write(_str)
        return _str

    def run(self) -> float:
        """
        Run the random search.

        In each iteration, new parameters are sampled, a training is performed
        and evaluated. Training files are copied into a directory named after
        the iteration.

        Returns
        -------
        float
            the best score
        """
        for i in range(self.current_run, self.current_run + self.iterations):
            self.logger.info(f"Train iteration {i}")

            seed = init_random_seed()
            self.logger.info(f"Random seed for iteration {i} is {seed}")

            self.setup_run()
            self.training.run()

            score = self.eval_run()
            self.logger.info(f"Iteration {i} has score {score:.6f}")

            self.current_score = min(score, self.current_score)
            self.cleanup_run(i, link_best=(self.current_score == score))
        return self.current_score

    def setup_run(self):
        """
        Setup a run.

        This methods performs sampling, creates necessary directories and
        delegates the setup of a training to the `setup_training` method.

        Parameters
        """
        self.logger.debug(f"Create directories training directory {self.train_dir}")
        os.mkdir(self.train_dir)
        snapshot_dir = os.path.join(self.train_dir, "snapshots")
        os.mkdir(snapshot_dir)

        self.params = self.cleanup_params(self.sample())
        params_file = os.path.join(self.train_dir, "params.json")
        self.serialize_params(self.params, params_file)
        self.logger.debug(f"Store params at {params_file}")

        logfile = os.path.join(self.train_dir, "train.log")
        logger = get_logger(logfile, loglevel="INFO", name="Training")
        self.logger.debug(f"Write training logs to {logfile}")

        self.training = self.setup_training(
            self.params, self.train_dir, snapshot_dir, logger
        )

    def cleanup_run(self, iteration: int, link_best: bool = False):
        """
        Cleanup of a run.

        All training data are copied to a direcotry named after the iteration number
        and optionally the new directory is linked as the best run.

        Parameters
        ----------
        iteration: int
            the current iteration of random search
        link_best: bool
            if the current run is the best performing one
        """
        print(iteration)
        run_dir = os.path.join(self.out_dir, f"{iteration:03d}")
        self.logger.debug(
            f"Move training data of iteration {iteration} from {self.train_dir} to {run_dir}"
        )
        shutil.move(self.train_dir, run_dir)

        if link_best:
            linkfile = os.path.join(self.out_dir, "best")
            if os.path.exists(linkfile):
                os.unlink(linkfile)
            os.symlink(os.path.basename(run_dir), linkfile)

    def setup_training(
        self, params: Dict[str, Any], train_dir: str, snapshot_dir: str, logger: Logger
    ) -> AbstractTraining:
        """
        Setup a training.

        Parameters
        ----------
        params: Dict[str, Any]
            the sampled parameters for the run
        train_dir: str
            the directory where the training should store its data
        snapshot_dir: str
            the directory where the training should store its snapshots (usually a sub-directory of train_dir)
        logger: Logger
        """
        pass

    def eval_run(self) -> float:
        """
        Evaluate a training.

        Returns
        -------
        float
            the score of the training
        """
        pass

    def get_score(self, train_dir: str) -> float:
        """
        Get the score of an already completed random search run by its directory.

        Parameters
        ----------
        train_dir: str
            directory of the random search run results

        Returns
        -------
        float
            the score of the given run
        """
        pass

    def cleanup_params(self, params: Dict[str, Any]) -> Dict[str, Any]:
        """
        Method can be implemented to perform some cleanup on parameters
        after sampling.

        Parameters
        ----------
        params: Dict[str, Any]
            the currently sampled params

        Returns
        -------
        Dict[str, Any]
        """
        return params
